# -*- coding: utf-8 -*-
"""IPython extension for a mcpyrate-enabled REPL.

To enable::

    %load_ext mcpyrate.repl.iconsole

To autoload it at IPython startup, put this into your ``ipython_config.py``::

    c.InteractiveShellApp.extensions = ["mcpyrate.repl.iconsole"]

To find your config file, ``ipython profile locate``.

The line magic `%macros` shows macros currently imported to the session.

The cell magic `%%dump_ast` pretty-prints the AST of a cell. But note that
you can also use the macro `mcpyrate.debug.step_expansion` in the REPL.

The function `macro(f)` binds a `f` as a macro. Works also as a decorator.
Note this comes from `mcpyrate`; it's different from IPython's `%macro`
line magic.
"""

import ast
import copy
from functools import partial
import sys

from IPython.core import magic_arguments
from IPython.core.error import InputRejected
from IPython.core.magic import Magics, magics_class, cell_magic, line_magic

from .. import __version__ as mcpyrate_version
from ..astdumper import dump
from ..compiler import create_module
from ..debug import format_bindings
from ..expander import find_macros, MacroExpander, global_postprocess
from .utils import get_makemacro_sourcecode

# Boot up `mcpyrate` so that the REPL can import modules that use macros.
# Despite the meta-levels, there's just one global importer for the Python process.
from .. import activate  # noqa: F401

_magic_module_name = "__mcpyrate_repl_self__"
_placeholder = "<ipython-session>"
_instance = None

def load_ipython_extension(ipython):
    # TODO: The banner is injected too late.
    #
    # It seems IPython startup is
    # already complete when ``load_ipython_extension()`` is called.
    #
    # We shouldn't print anything directly here; doing that breaks tools
    # such as the Emacs Python autoimporter (see importmagic.el in Spacemacs;
    # it will think epc failed to start if anything but the bare process id
    # is printed). Tools expect to suppress **all** of the IPython banner
    # by telling IPython itself not to print it.
    #
    # For now, let's just put the info into banner2, and refrain from printing it.
    # https://stackoverflow.com/questions/31613804/how-can-i-call-ipython-start-ipython-with-my-own-banner
    ipython.config.TerminalInteractiveShell.banner2 = "mcpyrate {} -- Advanced macro expander for Python.".format(mcpyrate_version)
    global _instance
    if not _instance:
        _instance = IMcpyrateExtension(shell=ipython)
        ipython.register_magics(AstMagics)

# TODO: unregister magics at unload time?
def unload_ipython_extension(ipython):
    global _instance
    _instance = None


@magics_class
class AstMagics(Magics):
    """IPython magics related to `mcpyrate` macro support."""

    # avoid complaining about typoed macro names when the macro functions are loaded
    @cell_magic
    def ignore_importerror(self, line, cell):
        """Run `cell`, ignoring any `ImportError` silently."""
        try:
            # Set globals to the shell user namespace to respect assignments
            # made by code in the cell. (`import` is a binding construct!)
            exec(cell, self.shell.user_ns)
        except ImportError:
            pass

    @line_magic
    def macros(self, line):
        """Print a human-readable list of macros currently imported into the session."""
        # Line magics print an extra `\n` at the end automatically, so remove our final `\n`.
        print(format_bindings(_instance.macro_transformer.expander, color=True), end="")

    # I don't know if this is useful - one can use the `mcpyrate.debug.step_expansion`
    # macro also in the REPL - but let's put it in for now.
    # http://alexleone.blogspot.co.uk/2010/01/python-ast-pretty-printer.html
    @magic_arguments.magic_arguments()
    @magic_arguments.argument(
        "-m", "--mode", default="exec",
        help="The mode in which to parse the code. Can be exec (default), "
             "eval or single."
    )
    # TODO: add support for expand-once
    @magic_arguments.argument(
        "-e", "--expand", default="no",
        help="Whether to expand macros before dumping the AST. Can be yes "
             "or no (default)."
    )
    @cell_magic
    def dump_ast(self, line, cell):
        """Parse the code in the cell, and pretty-print the AST."""
        args = magic_arguments.parse_argstring(self.dump_ast, line)
        tree = ast.parse(cell, mode=args.mode)
        if args.expand != "no":
            tree = _instance.macro_transformer.visit(tree)
        print(dump(tree))


class InteractiveMacroTransformer(ast.NodeTransformer):
    """AST transformer for IPython that expands `mcpyrate` macros."""

    def __init__(self, extension_instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ipyextension = extension_instance
        self.expander = MacroExpander(bindings={}, filename="<ipython-session>")

    def visit(self, tree):
        try:
            sys.modules[_magic_module_name].__dict__.clear()
            sys.modules[_magic_module_name].__dict__.update(self._ipyextension.shell.user_ns)  # for self-macro-imports
            # We treat the initial magic module metadata as write-protected: even if the user
            # defines a variable of the same name in the user namespace, the metadata fields
            # in the magic module won't be overwritten. (IPython actually defines e.g. `__name__`
            # in `user_ns` even if the user explicitly doesn't.)
            sys.modules[_magic_module_name].__dict__.update(self._ipyextension.magic_module_metadata)
            # macro-imports (this will import the modules)
            bindings = find_macros(tree, filename=self.expander.filename,
                                   reload=True, self_module=_magic_module_name)
            if bindings:
                self._ipyextension._macro_bindings_changed = True
                self.expander.bindings.update(bindings)
            new_tree = self.expander.visit(tree)
            new_tree = global_postprocess(new_tree)
            self._ipyextension.src = _placeholder
            return new_tree
        except Exception as err:
            # see IPython.core.interactiveshell.InteractiveShell.transform_ast()
            raise InputRejected(*err.args)


class IMcpyrateExtension:
    """IPython extension. Import, define and use `mcpyrate` macros in the IPython REPL."""

    def __init__(self, shell):
        self._macro_bindings_changed = False
        self.src = _placeholder
        self.shell = shell

        self.shell.input_transformers_post.append(self._get_source_code)

        self.macro_transformer = InteractiveMacroTransformer(extension_instance=self)
        self.shell.ast_transformers.append(self.macro_transformer)  # TODO: last or first?
        # Lucky that both meta-levels speak the same language, eh?
        shell.user_ns["__macro_expander__"] = self.macro_transformer.expander

        # support `from __self__ import macros, ...`
        magic_module = create_module(dotted_name=_magic_module_name, filename=_placeholder)
        self.magic_module_metadata = copy.copy(magic_module.__dict__)

        self.shell.run_cell(get_makemacro_sourcecode(),
                            store_history=False,
                            silent=True)

        # TODO: If we want to support dialects in the REPL, we need to install
        # a string transformer here to call the dialect system's source transformer,
        # and then modify `InteractiveMacroTransformer` to run the dialect system's
        # AST transformer before it runs the macro expander. Look at `mcpyrate.compiler.compile`.

        ipy = self.shell.get_ipython()
        ipy.events.register("post_run_cell", self._refresh_macro_functions)

    def __del__(self):
        ipy = self.shell.get_ipython()
        ipy.events.unregister("post_run_cell", self._refresh_macro_functions)
        del self.shell.user_ns["__macro_expander__"]
        self.shell.ast_transformers.remove(self.macro_transformer)
        self.shell.input_transformers_post.remove(self._get_source_code)

    def _get_source_code(self, lines):  # IPython 7.0+ with Python 3.5+
        """Get the source code of the current cell.

        This is a do-nothing string transformer that just captures the text.
        It is intended to run last, just before any AST transformers run.
        """
        self.src = lines
        return lines

    def _refresh_macro_functions(self, info):
        """Refresh macro function imports.

        Called after running a cell, so that IPython help "some_macro?" works
        for the currently available macros, allowing the user to easily view
        macro docstrings.
        """
        if not self._macro_bindings_changed:
            return
        self._macro_bindings_changed = False
        internal_execute = partial(self.shell.run_cell,
                                   store_history=False,
                                   silent=True)

        for asname, function in self.macro_transformer.expander.bindings.items():
            # Catch broken bindings due to erroneous imports in user code
            # (e.g. accidentally to a module object instead of to a function object)
            if not (hasattr(function, "__module__") and hasattr(function, "__qualname__")):
                continue
            if not function.__module__:  # Macros defined in the REPL have `__module__=None`.
                continue
            commands = ["%%ignore_importerror",
                        f"from {function.__module__} import {function.__qualname__} as {asname}"]
            internal_execute("\n".join(commands))



from .astdumper import dump  # noqa: F401
from .core import MacroExpansionError  # noqa: F401
from .dialects import Dialect  # noqa: F401
from .expander import namemacro, parametricmacro  # noqa: F401
from .unparser import unparse  # noqa: F401
from .utils import gensym  # noqa: F401

__version__ = "3.4.0"



# -*- coding: utf-8 -*-
"""mcpyrate-enabled `code.InteractiveConsole`.

Special commands:

  - `obj?` shows obj's docstring, `obj??` shows its source code.
  - `macros?` shows macro bindings.
  - `macro(f)` binds a function as a macro. Works also as a decorator.
"""

# Based on `imacropy.console.MacroConsole` by Juha Jeronen,
# which was based on `macropy.core.MacroConsole` by Li Haoyi,
# Justin Holmgren, Alberto Berti and all the other contributors,
# 2013-2019. Used under the MIT license.
#    https://github.com/azazel75/macropy
#    https://github.com/Technologicat/imacropy

__all__ = ["MacroConsole"]

import ast
import code
import copy
import sys
import textwrap

from .. import __version__ as mcpyrate_version
from ..compiler import create_module
from ..core import MacroExpansionError
from ..debug import format_bindings
from ..expander import find_macros, MacroExpander, global_postprocess
from .utils import get_makemacro_sourcecode

# Boot up `mcpyrate` so that the REPL can import modules that use macros.
# Despite the meta-levels, there's just one global importer for the Python process.
from .. import activate  # noqa: F401

_magic_module_name = "__mcpyrate_repl_self__"

class MacroConsole(code.InteractiveConsole):
    def __init__(self, locals=None, filename="<interactive input>"):
        """Parameters like in `code.InteractiveConsole`."""
        self.expander = MacroExpander(bindings={}, filename=filename)
        self._macro_bindings_changed = False

        if locals is None:
            locals = {}
        # Lucky that both meta-levels speak the same language, eh?
        locals["__macro_expander__"] = self.expander

        # support `from __self__ import macros, ...`
        #
        # To do this, we need the top-level variables in the REPL to be stored
        # in the namespace of a module, so that `find_macros` can look them up.
        #
        # We create a module dynamically. The one obvious way would be to then
        # replace the `__dict__` of that module with the given `locals`, but
        # `types.ModuleType` is special in that the `__dict__` binding itself
        # is read-only:
        #   https://docs.python.org/3/library/stdtypes.html#modules
        #
        # This leaves two possible solutions:
        #
        #   a. Keep our magic module separate from `locals`, and shallow-copy
        #      the user namespace into its `__dict__` after each REPL input.
        #      The magic module's namespace is only used for macro lookups.
        #
        #      + Behaves exactly like `code.InteractiveConsole` in that the
        #        user-given `locals` *is* the dict where the top-level variables
        #        are stored. So the user can keep a reference to that dict, and
        #        they will see any changes made by assigning to top-level
        #        variables in the REPL.
        #
        #      - May slow down the REPL when a lot of top-level variables exist.
        #        Likely not a problem in practice.
        #
        #   b. Use the module's `__dict__` as the local namespace of the REPL
        #      session. At startup, update it from `locals`, to install any
        #      user-provided initial variable bindings for the session.
        #
        #      + Elegant. No extra copying.
        #
        #      - Behavior does not match `code.InteractiveConsole`. The
        #        `locals` argument only provides initial values; any updates
        #        made during the REPL session are stored in the module's
        #        `__dict__`, which is a different dict instance. Hence the user
        #        will not see any changes made by assigning to top-level
        #        variables in the REPL.
        #
        # We currently use strategy a., for least astonishment.
        #
        magic_module = create_module(dotted_name=_magic_module_name, filename=filename)
        self.magic_module_metadata = copy.copy(magic_module.__dict__)

        super().__init__(locals, filename)

        # Support for special REPL commands.
        self._internal_execute(get_makemacro_sourcecode())
        self._internal_execute("import mcpyrate.repl.utils")

    def _internal_execute(self, source):
        """Execute given source in the console session.

        This is support magic for internal operation of the console
        session itself, e.g. for auto-loading macro functions.

        The source must be pure Python, i.e. no macros.

        The source is NOT added to the session history.

        This bypasses `runsource`, so it too can use this function.
        """
        source = textwrap.dedent(source)
        tree = ast.parse(source)
        tree = ast.Interactive(tree.body)
        code = compile(tree, "<console internal>", "single", self.compile.compiler.flags, 1)
        self.runcode(code)

    def interact(self, banner=None, exitmsg=None):
        """See `code.InteractiveConsole.interact`.

        The only thing we customize here is that if `banner is None`, in which case
        `code.InteractiveConsole` will print its default banner, we print help for
        our special commands and a line containing the `mcpyrate` version before that
        default banner.
        """
        if banner is None:
            self.write(f"mcpyrate {mcpyrate_version} -- Advanced macro expander for Python.\n")
            self.write("- obj? to view obj's docstring, and obj?? to view its source code.\n")
            self.write("- macros? to view macro bindings.\n")
            self.write("- macro(f) to bind function f as a macro. Works also as a decorator.\n")
        return super().interact(banner, exitmsg)

    def runsource(self, source, filename="<interactive input>", symbol="single"):
        # Special REPL commands.
        if source == "macros?":
            self.write(format_bindings(self.expander, color=True))
            return False  # complete input
        elif source.endswith("??"):
            # Use `_internal_execute` instead of `runsource` to prevent expansion of name macros.
            return self._internal_execute(f"mcpyrate.repl.utils.sourcecode({source[:-2]})")
        elif source.endswith("?"):
            return self._internal_execute(f"mcpyrate.repl.utils.doc({source[:-1]})")

        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            code = ""
        if code is None:  # incomplete input
            return True

        try:
            # TODO: If we want to support dialects in the REPL, this is where to do it.
            #       Look at `mcpyrate.compiler.compile`.
            tree = ast.parse(source)

            # macro-imports (this will import the modules)
            sys.modules[_magic_module_name].__dict__.clear()
            sys.modules[_magic_module_name].__dict__.update(self.locals)  # for self-macro-imports
            # We treat the initial magic module metadata as write-protected: even if the user
            # defines a variable of the same name in the user namespace, the metadata fields
            # in the magic module won't be overwritten.
            sys.modules[_magic_module_name].__dict__.update(self.magic_module_metadata)
            bindings = find_macros(tree, filename=self.expander.filename,
                                   reload=True, self_module=_magic_module_name)
            if bindings:
                self._macro_bindings_changed = True
                self.expander.bindings.update(bindings)
            tree = self.expander.visit(tree)
            tree = global_postprocess(tree)

            tree = ast.Interactive(tree.body)
            code = compile(tree, filename, symbol, self.compile.compiler.flags, 1)
        except (OverflowError, SyntaxError, ValueError, MacroExpansionError):
            self.showsyntaxerror(filename)
            return False  # erroneous input
        except ModuleNotFoundError as err:  # during macro module lookup
            # In this case, the standard stack trace is long and points only to our code and the stdlib,
            # not the erroneous input that's the actual culprit. Better ignore it, and emulate showsyntaxerror.
            # TODO: support sys.excepthook.
            # TODO: Look at `code.InteractiveConsole.showsyntaxerror` for how to do that.
            self.write(f"{err.__class__.__name__}: {str(err)}\n")
            return False  # erroneous input
        except ImportError as err:  # during macro lookup in a successfully imported module
            self.write(f"{err.__class__.__name__}: {str(err)}\n")
            return False  # erroneous input

        self.runcode(code)
        self._refresh_macro_functions()
        return False  # Successfully compiled. `runcode` takes care of any runtime failures.

    def _refresh_macro_functions(self):
        """Refresh macro function imports.

        Called after successfully compiling and running an input, so that
        `some_macro.__doc__` points to the right docstring.
        """
        if not self._macro_bindings_changed:
            return
        self._macro_bindings_changed = False

        for asname, function in self.expander.bindings.items():
            # Catch broken bindings due to erroneous imports in user code
            # (e.g. accidentally to a module object instead of to a function object)
            if not (hasattr(function, "__module__") and hasattr(function, "__qualname__")):
                continue
            if not function.__module__:    # Macros defined in the REPL have `__module__=None`.
                continue
            try:
                source = f"from {function.__module__} import {function.__qualname__} as {asname}"
                self._internal_execute(source)
            except (ModuleNotFoundError, ImportError):
                pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Universal bootstrapper for macro-enabled Python programs powered by mcpyrate."""

# TODO: Currently tested in CPython 3.6, and PyPy3 7.3.0 (Python 3.6). Test in 3.7+.

import argparse
import atexit
import os
import pathlib
import sys
from importlib import import_module
from importlib.util import module_from_spec, resolve_name

from .. import __version__ as __mcpyrate_version__
from .. import activate  # noqa: F401
from ..core import MacroApplicationError
from ..coreutils import relativize
from ..pycachecleaner import deletepycachedirs, getpycachedirs

__version__ = "3.1.0"

_config_dir = "~/.config/mcpyrate"
_macropython_module = None  # sys.modules doesn't always seem to keep it, so stash it locally too.

# --------------------------------------------------------------------------------

def import_module_as_main(name, script_mode):
    """Import a module, pretending it's __main__.

    We support only new-style loaders that provide `exec_module`; old-style
    loaders that provide `load_module` and `create_module` are not supported.

    Upon success, replaces `sys.modules["__main__"]` with the module that
    was imported. Upon failure, propagates any exception raised.

    This is a customized approximation of the standard import semantics, based on:

        https://docs.python.org/3/library/importlib.html#approximating-importlib-import-module
        https://docs.python.org/3/reference/import.html#loading
        https://docs.python.org/3/reference/import.html#import-related-module-attributes
        https://docs.python.org/3/reference/import.html#module-path
        https://docs.python.org/3/reference/import.html#special-considerations-for-main
    """
    absolute_name = resolve_name(name, package=None)

    # If already loaded, normally we should return the module from `sys.modules`,
    # but `sys.modules["__main__"]` is for now this bootstrapper, not the user __main__.
    #
    # Note __main__ and somemod are distinct modules, even when the same file,
    # because __main__ triggers `if __name__ == '__main__'` checks, but somemod doesn't.
    #
    # try:
    #     return sys.modules[absolute_name]
    # except KeyError:
    #     pass

    path = None
    if '.' in absolute_name:
        if script_mode:
            raise ValueError("In script mode, please add the containing directory to `sys.path` and then top-level import the final component of the name.")
        # Import the module's parent package normally to initialize parent packages.
        # Get the appropriate `path` for `find_spec` to find subpackages and modules.
        parent_name, _, child_name = absolute_name.rpartition('.')
        parent_module = import_module(parent_name)
        path = parent_module.__spec__.submodule_search_locations

    for finder in sys.meta_path:
        if not hasattr(finder, "find_spec"):  # pkg_resources.extern.VendorImporter has no find_spec
            continue
        # https://docs.python.org/3/library/importlib.html#importlib.abc.MetaPathFinder.find_spec
        spec = finder.find_spec(absolute_name, path)
        if spec is not None:
            break
    else:
        msg = f"No module named {absolute_name}"
        raise ModuleNotFoundError(msg, name=absolute_name)

    spec.name = "__main__"
    if spec.loader:
        spec.loader.name = "__main__"

    module = module_from_spec(spec)
    try_mainpy = False
    if script_mode:  # e.g. "macropython somepackage/__init__.py"
        module.__package__ = ""
    elif path:  # e.g. "macropython -m somepackage.module"
        module.__package__ = parent_name
    elif spec.origin.endswith("__init__.py"):  # e.g. "macropython -m somepackage"
        module.__package__ = absolute_name
        try_mainpy = True

    if spec.origin == "namespace":
        # https://docs.python.org/3/reference/import.html#__path__
        module.__path__ = spec.submodule_search_locations

    if try_mainpy:
        # e.g. "import somepackage" in the above case; it's not __main__, so import it normally.
        assert not script_mode
        parent_module = import_module(absolute_name)
    elif spec.loader:
        sys.modules["__main__"] = module  # replace this bootstrapper with the new __main__
        try:
            spec.loader.exec_module(module)
        except Exception as err:
            sys.modules["__main__"] = _macropython_module
            if isinstance(err, MacroApplicationError):
                # To avoid noise, discard most of the traceback of the chained
                # macro-expansion errors emitted by the expander core. The
                # linked (`__cause__`) exceptions have the actual tracebacks.
                #
                # Keep just the last entry, which should state that this
                # exception came from `expand` in `core.py`.
                tb = err.__traceback__
                while tb.tb_next:
                    tb = tb.tb_next
                raise err.with_traceback(tb)
            raise
        # # __main__ has no parent module so we don't need to do this.
        # if path is not None:
        #     setattr(parent_module, child_name, module)
    else:  # namespace packages have loader=None
        try_mainpy = True

    # __init__.py (if any) has run; run __main__.py, like `python -m somepackage` does.
    if try_mainpy:
        has_mainpy = True
        try:
            # __main__.py doesn't need to have its name set to "__main__", so import it normally.
            import_module(f"{absolute_name}.__main__")
        except ImportError as e:
            if "No module named" in e.msg:
                has_mainpy = False
            else:
                raise
        if not has_mainpy:
            raise ImportError(f"No module named {absolute_name}.__main__; '{absolute_name}' is a package and cannot be directly executed")

    return module

def main():
    """Handle command-line arguments and run the specified main program."""
    parser = argparse.ArgumentParser(description="""Run a Python program or an interactive interpreter with mcpyrate enabled.""",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-v', '--version', action='version', version=('%(prog)s ' + __version__ + ' (mcpyrate ' + __mcpyrate_version__ + ')'))
    parser.add_argument(dest='filename', nargs='?', default=None, type=str, metavar='file',
                        help='script to run')
    parser.add_argument('-m', '--module', dest='module', default=None, type=str, metavar='mod',
                        help='run library module as a script (like python -m mod)')
    parser.add_argument('-i', '--interactive', dest='interactive', action="store_true", default=False,
                        help='interactive mode (macro-enabled Python interpreter). '
                             'If -i is given, <filename> and -m <mod> are ignored.')
    parser.add_argument('-p', '--pylab', dest='pylab', action="store_true", default=False,
                        help='For use together with "-i". Automatically "import numpy as np", '
                             '"import matplotlib.pyplot as plt", and enable mpl\'s interactive '
                             'mode (somewhat like IPython\'s pylab mode).')
    parser.add_argument('-c', '--clean', dest='path_to_clean', default=None, type=str, metavar='dir',
                        help='Delete Python bytecode (`.pyc`) caches inside given directory, recursively, '
                             'and then exit. This removes the `__pycache__` directories, too. '
                             'The purpose is to facilitate testing `mcpyrate` and programs that use it, '
                             'because the expander only runs when the `.pyc` cache for the module being '
                             'imported is out of date or does not exist. For any given module, once the '
                             'expander is done, Python\'s import system automatically writes the `.pyc` '
                             'cache for that module.')
    parser.add_argument("-n", "--dry-run", dest="dry_run", action="store_true", default=False,
                        help='For use together with "-c". Just scan for and print the .pyc cache '
                             'directory paths, don\'t actually clean them.')
    opts = parser.parse_args()

    if opts.path_to_clean:
        # If an error occurs during cleaning, we just let it produce a standard stack trace.
        if not opts.dry_run:
            deletepycachedirs(opts.path_to_clean)
        else:
            for x in getpycachedirs(opts.path_to_clean):
                print(x)
        sys.exit(0)  # only reached if cleaning (or dry run) successful

    if opts.interactive:
        import readline  # noqa: F401, side effect: enable GNU readline in input()
        import rlcompleter  # noqa: F401, side effects: readline tab completion

        from .console import MacroConsole
        repl_locals = {}
        if opts.pylab:  # like IPython's pylab mode, but we keep things in separate namespaces.
            import matplotlib.pyplot
            import numpy
            repl_locals["np"] = numpy
            repl_locals["plt"] = matplotlib.pyplot
            matplotlib.pyplot.ion()
        readline.set_completer(rlcompleter.Completer(namespace=repl_locals).complete)
        readline.parse_and_bind("tab: complete")  # PyPy ignores this, but not needed there.

        config_dir = pathlib.Path(_config_dir).expanduser().resolve()
        try:
            readline.read_history_file(config_dir / "macropython_history")
        except FileNotFoundError:
            pass

        def save_history():
            config_dir.mkdir(parents=True, exist_ok=True)
            readline.set_history_length(1000)
            readline.write_history_file(config_dir / "macropython_history")
        atexit.register(save_history)

        # Add CWD to import path like the builtin interactive console does.
        if sys.path[0] != "":
            sys.path.insert(0, "")
        m = MacroConsole(locals=repl_locals)
        return m.interact()

    if not opts.filename and not opts.module:
        parser.print_help()
        sys.exit(0)
    if opts.filename and opts.module:
        raise ValueError("Please specify just one main program to run (either filename.py or -m module, not both).")

    # Import the module, pretending its __name__ is "__main__".
    #
    # We must import so that macros get expanded, so we can't use
    # runpy.run_module here (which just execs without importing).
    #
    # As for sys.path, we're not interested in the directory the `macropython`
    # script itself lives in. To imitate how Python itself behaves:
    #
    #   - If run as `macropython -m something`, then `sys.path[0]`
    #     should be set to the cwd.
    #   - If run as `macropython some/path/to/script.py`, then `sys.path[0]`
    #     should be set to the directory containing `script.py`.
    #
    if opts.module:
        # Python 3.7 and later resolve the cwd at interpreter startup time, and use that.
        # Python 3.6 and earlier use the empty string, which dynamically resolves to
        # whatever is the current cwd whenever a module is imported.
        #     https://bugs.python.org/issue33053
        if sys.version_info >= (3, 7, 0):
            cwd = str(pathlib.Path.cwd().expanduser().resolve())
            if sys.path[0] != cwd:
                sys.path.insert(0, cwd)
        else:
            if sys.path[0] != "":
                sys.path.insert(0, "")

        # https://www.python.org/dev/peps/pep-0582/
        if sys.version_info >= (3, 8, 0):
            local_pypackages_dir = pathlib.Path.cwd().expanduser().resolve() / "__pypackages__"
            if local_pypackages_dir.is_dir():
                # TODO: figure out correct insert index (spec: "after cwd, just before site packages")
                sys.path.insert(1, str(local_pypackages_dir))

        import_module_as_main(opts.module, script_mode=False)

    else:  # no module, so use opts.filename
        fullpath = pathlib.Path(opts.filename).expanduser().resolve()
        if not fullpath.is_file():
            raise FileNotFoundError(f"Can't open file '{opts.filename}'")

        containing_directory = str(fullpath.parent)
        if sys.path[0] != containing_directory:
            sys.path.insert(0, containing_directory)

        # https://www.python.org/dev/peps/pep-0582/
        if sys.version_info >= (3, 8, 0):
            local_pypackages_dir = fullpath.parent / "__pypackages__"
            if local_pypackages_dir.is_dir():
                # TODO: figure out correct insert index (spec: "after cwd, just before site packages")
                sys.path.insert(1, str(local_pypackages_dir))

        # This approach finds the standard loader even for macro-enabled scripts.
        # TODO: For mcpyrate, that could be ok, since we monkey-patch just that.
        # TODO: But maybe better to leave the option to replace the whole loader later.
        # spec = spec_from_file_location("__main__", str(fullpath))
        # if not spec:
        #     raise ImportError(f"Not a Python module: '{opts.filename}'"
        # module = module_from_spec(spec)
        # TODO: if we use this approach, we should first initialize parent packages.
        # sys.modules[module.__name__] = module
        # spec.loader.exec_module(module)

        root_path, relative_path = relativize(opts.filename)
        module_name = relative_path.replace(os.path.sep, '.')
        if module_name.endswith(".__init__.py"):
            module_name = module_name[:-12]
        elif module_name.endswith(".py"):
            module_name = module_name[:-3]

        import_module_as_main(module_name, script_mode=True)

if __name__ == "__main__":
    _macropython_module = sys.modules["__macropython__"] = sys.modules["__main__"]
    main()



#!/usr/bin/env python

import ast
import os

from setuptools import setup  # type: ignore[import]


def read(*relpath, **kwargs):
    with open(os.path.join(os.path.dirname(__file__), *relpath),
              encoding=kwargs.get("encoding", "utf8")) as fh:
        return fh.read()

# Extract __version__ from the package __init__.py
# (since it's not a good idea to actually run __init__.py during the build process).
#
# http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
init_py_path = os.path.join("mcpyrate", "__init__.py")
version = None
try:
    with open(init_py_path) as f:
        for line in f:
            if line.startswith("__version__"):
                module = ast.parse(line)
                expr = module.body[0]
                assert isinstance(expr, ast.Assign)
                v = expr.value
                if type(v) is ast.Constant:
                    # mypy understands `isinstance(..., ...)` but not `type(...) is ...`,
                    # and we want to match on the exact type, not any subclass that might be
                    # added in some future Python version.
                    assert isinstance(v, ast.Constant)
                    version = v.value
                elif type(v) is ast.Str:  # TODO: Python 3.8: remove ast.Str
                    assert isinstance(v, ast.Str)  # mypy
                    version = v.s
                break
except FileNotFoundError:
    pass
if not version:
    raise RuntimeError(f"Version information not found in {init_py_path}")

setup(
    name="mcpyrate",
    version=version,
    packages=["mcpyrate", "mcpyrate.repl"],
    provides=["mcpyrate"],
    keywords=["macros", "syntactic-macros", "macro-expander", "metaprogramming", "import", "importer"],
    install_requires=["colorama>=0.4.4"],
    python_requires=">=3.6",
    author="Juha Jeronen and Salvador de la Puente González",
    author_email="juha.m.jeronen@gmail.com",
    url="https://github.com/Technologicat/mcpyrate",
    description="Advanced macro expander and language lab for Python",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="MIT",
    platforms=["Any"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Utilities"
    ],
    entry_points={"console_scripts": ["macropython=mcpyrate.repl.macropython:main"]},
    zip_safe=False  # macros are not zip safe, because the zip importer fails to find sources.
)



# -*- coding: utf-8; -*-
"""Utilities for building REPLs."""

__all__ = ["doc", "sourcecode", "get_makemacro_sourcecode"]

from functools import partial
import inspect
import sys
import textwrap

from ..colorizer import colorize, ColorScheme


def _get_source(obj):
    # `inspect.getsourcefile` accepts "a module, class, method, function,
    # traceback, frame, or code object" (the error message says this if
    # we try it on something invalid).
    #
    # So if `obj` is an instance, we need to try again with its `__class__`.
    for x in (obj, obj.__class__):  # TODO: other places to fall back to?
        try:
            filename = inspect.getsourcefile(x)
            source, firstlineno = inspect.getsourcelines(x)
            return filename, source, firstlineno
        except (TypeError, OSError):
            continue
    raise NotImplementedError


def doc(obj, *, file=None, end="\n"):
    """Print an object's docstring non-interactively.

    If available, print also the filename and the starting line number
    of the definition of `obj`.

    The default is to print to `sys.stderr` (resolved at call time, to respect
    possible overrides); to override, use the `file` parameter, which works
    like in the builtin `print`.

    The `end` parameter is passed to `print` as-is.
    """
    file = file or sys.stderr
    printer = partial(print, file=file, end=end)
    try:
        filename, source, firstlineno = _get_source(obj)
        printer(colorize(f"{filename}:{firstlineno}", ColorScheme.SOURCEFILENAME))
    except NotImplementedError:
        pass
    if not hasattr(obj, "__doc__") or not obj.__doc__:
        printer(colorize("<no docstring>", ColorScheme.GREYEDOUT))
        return
    print(inspect.cleandoc(obj.__doc__), file=file)


def sourcecode(obj, *, file=None, end=None):
    """Print an object's source code to `sys.stderr`, non-interactively.

    If available, print also the filename and the starting line number
    of the definition of `obj`.

    Default is to print to `sys.stderr` (resolved at call time, to respect
    possible overrides); to override, use the `file` argument, which works
    like in the builtin `print`.

    The `end` parameter is passed to `print` as-is.
    """
    file = file or sys.stderr
    printer = partial(print, file=file, end=end)
    try:
        filename, source, firstlineno = _get_source(obj)
        printer(colorize(f"{filename}:{firstlineno}", ColorScheme.SOURCEFILENAME))
        # TODO: No syntax highlighting for now, because we'd have to parse and unparse,
        # TODO: which loses the original formatting and comments.
        for line in source:
            printer(line.rstrip("\n"))
    except NotImplementedError:
        printer(colorize("<no source code available>", ColorScheme.GREYEDOUT))


def get_makemacro_sourcecode():
    """Return source code for the REPL's `macro` magic function.

    For injection by REPL into the session.

    We assume the expander instance has been bound to the global variable
    `__macro_expander__` inside the REPL session.
    """
    return textwrap.dedent('''
    def macro(function):
        """[mcpyrate] `macro(f)`: bind function `f` as a macro. Works also as a decorator. REPL only."""
        if not callable(function):
            raise TypeError(f"`function` must be callable, got {type(function)} with value {repr(function)}")
        if function.__name__ == "<lambda>":
            raise TypeError("`function` must be a named function, got a lambda.")
        __macro_expander__.bindings[function.__name__] = function
        return function
    ''')
